use core::mem;

use embassy_rp::{
    gpio::{Level, Output},
    spi::{
        Blocking, ClkPin, Config, CsPin, Instance as SpiInstance, MisoPin, MosiPin, Polarity, Spi,
    },
    Peripheral,
};
use embassy_time::{Duration, Timer};

#[allow(non_camel_case_types)]
pub struct PAW3212DB_TJDT<'a, L, CS, T>
where
    L: FnMut(&str),
    CS: CsPin<T> + 'a,
    T: SpiInstance,
{
    cs: Output<'a, CS>,
    spi: Spi<'a, T, Blocking>,

    pub log: L,
}

impl<'a, L, CS, T> PAW3212DB_TJDT<'a, L, CS, T>
where
    L: FnMut(&str),
    CS: CsPin<T> + 'a,
    T: SpiInstance,
{
    const READ: u8 = 0b0111_1111;
    const WRITE: u8 = 0b1000_0000;

    const REG_PRODUCT_ID1: u8 = 0x00;
    const REG_PRODUCT_ID2: u8 = 0x01;
    const REG_MOTION_STATUS: u8 = 0x02;
    const REG_DELTA_X: u8 = 0x03;
    const REG_DELTA_Y: u8 = 0x04;
    const REG_OPERATION_MODE: u8 = 0x05;
    const REG_CONFIGURATION: u8 = 0x06;
    const REG_WRITE_PROTECT: u8 = 0x09;
    const REG_SLEEP1: u8 = 0x0A;
    const REG_SLEEP2: u8 = 0x0B;
    const REG_SLEEP3: u8 = 0x0C;
    const REG_CPI_X: u8 = 0x0D;
    const REG_CPI_Y: u8 = 0x0E;
    const REG_DELTA_XY_HI: u8 = 0x12;
    const REG_IQC: u8 = 0x13;
    const REG_SHUTTER: u8 = 0x14;
    const REG_FRAME_AVG: u8 = 0x17;
    const REG_MOUSE_OPTION: u8 = 0x19;
    const REG_SPI_MODE: u8 = 0x26;

    pub async fn new(
        spi: impl Peripheral<P = T> + 'a,
        clk: impl Peripheral<P = impl ClkPin<T> + 'a> + 'a,
        mosi: impl Peripheral<P = impl MosiPin<T> + 'a> + 'a,
        miso: impl Peripheral<P = impl MisoPin<T> + 'a> + 'a,
        cs: impl Peripheral<P = CS> + 'a,
        log: L,
    ) -> PAW3212DB_TJDT<'a, L, CS, T> {
        let mut spi_config = Config::default();
        spi_config.frequency = 2_000_000;
        spi_config.polarity = Polarity::IdleHigh;

        let spi = Spi::new_blocking(spi, clk, mosi, miso, spi_config);

        let mut cs = Output::new(cs, Level::Low);

        cs.set_low();

        Timer::after(Duration::from_millis(2)).await;

        cs.set_high();

        Self { cs, spi, log }
    }

    async fn access<F, R>(&mut self, f: F) -> R
    where
        F: Fn(&mut Self) -> R,
    {
        self.cs.set_low();
        Timer::after(Duration::from_micros(1)).await;

        let res = f(self);

        self.cs.set_high();
        Timer::after(Duration::from_micros(2)).await;

        res
    }

    async fn read(&mut self, address: u8) -> u8 {
        loop {
            let res = self
                .access(move |me| {
                    if me.spi.blocking_write(&[address & Self::READ, 0]).is_err() {
                        (me.log)("send address (read) failed");
                    }

                    let mut buf = [0];

                    if me.spi.blocking_read(&mut buf).is_err() {
                        (me.log)("read failed");
                        return 0;
                    }

                    buf[0] << 1
                })
                .await;

            if res != address {
                return res;
            }
        }
    }

    async fn write(&mut self, address: u8, value: u8) {
        self.access(move |me| {
            if me
                .spi
                .blocking_write(&[address | Self::WRITE, value])
                .is_err()
            {
                (me.log)("send address (write) failed");
            }
        })
        .await;
    }
}

impl<'a, L, CS, T> PAW3212DB_TJDT<'a, L, CS, T>
where
    L: FnMut(&str),
    CS: CsPin<T> + 'a,
    T: SpiInstance,
{
    pub async fn delta_x(&mut self) -> i8 {
        unsafe { mem::transmute(self.read(Self::REG_DELTA_X).await) }
    }

    pub async fn delta_y(&mut self) -> i8 {
        unsafe { mem::transmute(self.read(Self::REG_DELTA_Y).await) }
    }

    pub async fn product_id_1(&mut self) -> u8 {
        self.read(Self::REG_PRODUCT_ID1).await
    }

    pub async fn product_id_2(&mut self) -> u8 {
        self.read(Self::REG_PRODUCT_ID2).await
    }

    pub async fn operation_mode(&mut self) -> u8 {
        self.read(Self::REG_OPERATION_MODE).await
    }

    pub async fn motion_status(&mut self) -> u8 {
        self.read(Self::REG_MOTION_STATUS).await
    }
}
